import * as toml from "toml";
import Resource from "./resource";
import {
  UrlProvider,
  ModrinthProvider,
  CurseForgeProvider,
  GitHubProvider,
} from "./provider";

export interface Metafile {
  name: string;
  filename: string;
  side: "server" | "client" | "both";
  provider: UrlProvider;
}

/**
 * Represents a file entry in the Packwiz index.
 */
export class IndexFileEntry {
  readonly file: Resource;
  readonly hash: string;
  readonly metafile: boolean;

  constructor(
    file: string | Resource,
    hash: string,
    metafile: boolean = false,
  ) {
    this.file = file instanceof Resource ? file : new Resource(file);
    this.hash = hash;
    this.metafile = metafile;
  }

  async parse(): Promise<Metafile> {
    if (this.metafile === false) {
      throw new Error(
        "Cannot parse non-metafiles! Use `file.fetchContents()` instead to get the file contents.",
      );
    }
    const fileContent = await this.file.fetchContents();
    const parsed = toml.parse(fileContent);

    const metafile: Metafile = {
      name: parsed.name,
      filename: parsed.filename,
      side: parsed.side,
      provider: this.parseProvider(parsed),
    };

    return metafile;
  }

  private parseProvider(parsed: any): UrlProvider {
    if (parsed.update?.modrinth) {
      return new ModrinthProvider(
        parsed.download.hash,
        parsed.download["hash-format"],
        new Resource(parsed.download.url),
        parsed.update.modrinth["mod-id"],
        parsed.update.modrinth["version"],
      );
    } else if (parsed.update?.curseforge) {
      return new CurseForgeProvider(
        parsed.download.hash,
        parsed.download["hash-format"],
        parsed.download.mode,
        parsed.update.curseforge["file-id"],
        parsed.update.curseforge["project-id"],
      );
    } else if (parsed.update?.github) {
      return new GitHubProvider(
        parsed.download.hash,
        parsed.download["hash-format"],
        new Resource(parsed.download.url),
        parsed.update.github.branch,
        parsed.update.github.regex,
        parsed.update.github.slug,
        parsed.update.github.tag,
      );
    } else if (parsed.download) {
      return new UrlProvider(
        parsed.download.hash,
        parsed.download["hash-format"],
        new Resource(parsed.download.url),
      );
    } else {
      throw new Error("Unknown provider in TOML.");
    }
  }
}

/**
 * Represents the structure of a Packwiz index file, as well as providing its location.
 */
export interface PackwizIndex {
  location: Resource;
  hashFormat: string;
  files: IndexFileEntry[];
}

// TODO: parse pack.toml too instead of just the index file
// This is because pack.toml contains important information like modloader versions

/**
 * Parses a Packwiz index file and returns its contents.
 *
 * @param indexFilePath - The path of the TOML file.
 * @returns The parsed index as a structured object.
 */
export async function parsePackwizIndex(
  indexFilePath: string,
): Promise<PackwizIndex> {
  const indexFile = new Resource(indexFilePath);
  const parsed = toml.parse(await indexFile.fetchContents());
  return {
    location: indexFile,
    hashFormat: parsed["hash-format"],
    files: parsed.files.map(
      (file: any) =>
        new IndexFileEntry(
          indexFile.parent.join(file.file),
          file.hash,
          file.metafile,
        ),
    ),
  };
}
